@@ -25,6 +25,10 @@
 #endif
 
 #include <libavcodec/avcodec.h>
+#include <libavutil/imgutils.h>
+#include <libavfilter/avfilter.h>
+#include <libavfilter/buffersrc.h>
+#include <libavfilter/buffersink.h>
 
 #include <gst/gst.h>
 #include <gst/video/video.h>
@@ -91,8 +95,16 @@ typedef struct _GstFFMpegDeinterlace
   gboolean reconfigure;
   GstFFMpegDeinterlaceMode new_mode;
 
-  enum PixelFormat pixfmt;
+  enum AVPixelFormat pixfmt;
   AVPicture from_frame, to_frame;
+
+  AVFilterContext *buffersink_ctx;
+  AVFilterContext *buffersrc_ctx;
+  AVFilterGraph *filter_graph;
+  AVFrame *filter_frame;
+  int last_width, last_height;
+  enum AVPixelFormat last_pixfmt;
+
 } GstFFMpegDeinterlace;
 
 typedef struct _GstFFMpegDeinterlaceClass
@@ -135,6 +147,8 @@ G_DEFINE_TYPE (GstFFMpegDeinterlace, gst_ffmpegdeinterlace, GST_TYPE_ELEMENT);
 static GstFlowReturn gst_ffmpegdeinterlace_chain (GstPad * pad,
     GstObject * parent, GstBuffer * inbuf);
 
+static void gst_ffmpegdeinterlace_dispose (GObject * obj);
+
 static void
 gst_ffmpegdeinterlace_class_init (GstFFMpegDeinterlaceClass * klass)
 {
@@ -167,6 +181,8 @@ gst_ffmpegdeinterlace_class_init (GstFFMpegDeinterlaceClass * klass)
   gst_element_class_set_static_metadata (element_class,
       "libav Deinterlace element", "Filter/Effect/Video/Deinterlace",
       "Deinterlace video", "Luca Ognibene <luogni@tin.it>");
+
+  gobject_class->dispose = gst_ffmpegdeinterlace_dispose;
 }
 
 static void
@@ -206,9 +222,9 @@ gst_ffmpegdeinterlace_sink_setcaps (GstPad * pad, GstCaps * caps)
   ctx = avcodec_alloc_context3 (NULL);
   ctx->width = deinterlace->width;
   ctx->height = deinterlace->height;
-  ctx->pix_fmt = PIX_FMT_NB;
+  ctx->pix_fmt = AV_PIX_FMT_NB;
   gst_ffmpeg_caps_with_codectype (AVMEDIA_TYPE_VIDEO, caps, ctx);
-  if (ctx->pix_fmt == PIX_FMT_NB) {
+  if (ctx->pix_fmt == AV_PIX_FMT_NB) {
     gst_ffmpeg_avcodec_close (ctx);
     av_free (ctx);
     return FALSE;
@@ -219,8 +235,8 @@ gst_ffmpegdeinterlace_sink_setcaps (GstPad * pad, GstCaps * caps)
   av_free (ctx);
 
   deinterlace->to_size =
-      avpicture_get_size (deinterlace->pixfmt, deinterlace->width,
-      deinterlace->height);
+      av_image_get_buffer_size (deinterlace->pixfmt, deinterlace->width,
+      deinterlace->height, 1);
 
   src_caps = gst_caps_copy (caps);
   gst_caps_set_simple (src_caps, "interlace-mode", G_TYPE_STRING,
@@ -270,13 +286,111 @@ gst_ffmpegdeinterlace_init (GstFFMpegDeinterlace * deinterlace)
   deinterlace->srcpad = gst_pad_new_from_static_template (&src_factory, "src");
   gst_element_add_pad (GST_ELEMENT (deinterlace), deinterlace->srcpad);
 
-  deinterlace->pixfmt = PIX_FMT_NB;
+  deinterlace->pixfmt = AV_PIX_FMT_NB;
 
   deinterlace->interlaced = FALSE;
   deinterlace->passthrough = FALSE;
   deinterlace->reconfigure = FALSE;
   deinterlace->mode = DEFAULT_MODE;
   deinterlace->new_mode = -1;
+  deinterlace->last_width = -1;
+  deinterlace->last_height = -1;
+  deinterlace->last_pixfmt = AV_PIX_FMT_NONE;
+}
+
+static void
+delete_filter_graph (GstFFMpegDeinterlace * deinterlace)
+{
+  if (deinterlace->filter_graph) {
+    av_frame_free (&deinterlace->filter_frame);
+    avfilter_graph_free (&deinterlace->filter_graph);
+  }
+}
+
+static void
+gst_ffmpegdeinterlace_dispose (GObject * obj)
+{
+  GstFFMpegDeinterlace *deinterlace = GST_FFMPEGDEINTERLACE (obj);
+
+  delete_filter_graph (deinterlace);
+
+  G_OBJECT_CLASS (gst_ffmpegdeinterlace_parent_class)->dispose (obj);
+}
+
+static int
+init_filter_graph (GstFFMpegDeinterlace * deinterlace,
+    enum AVPixelFormat pixfmt, int width, int height)
+{
+  AVFilterInOut *inputs = NULL, *outputs = NULL;
+  char args[512];
+  int res;
+
+  delete_filter_graph (deinterlace);
+  deinterlace->filter_graph = avfilter_graph_alloc ();
+  snprintf (args, sizeof (args),
+      "buffer=video_size=%dx%d:pix_fmt=%d:time_base=1/1:pixel_aspect=0/1[in];"
+      "[in]yadif[out];" "[out]buffersink", width, height, pixfmt);
+  res =
+      avfilter_graph_parse2 (deinterlace->filter_graph, args, &inputs,
+      &outputs);
+  if (res < 0)
+    return res;
+  if (inputs || outputs)
+    return -1;
+  res = avfilter_graph_config (deinterlace->filter_graph, NULL);
+  if (res < 0)
+    return res;
+
+  deinterlace->buffersrc_ctx =
+      avfilter_graph_get_filter (deinterlace->filter_graph, "Parsed_buffer_0");
+  deinterlace->buffersink_ctx =
+      avfilter_graph_get_filter (deinterlace->filter_graph,
+      "Parsed_buffersink_2");
+  if (!deinterlace->buffersrc_ctx || !deinterlace->buffersink_ctx)
+    return -1;
+  deinterlace->filter_frame = av_frame_alloc ();
+  deinterlace->last_width = width;
+  deinterlace->last_height = height;
+  deinterlace->last_pixfmt = pixfmt;
+
+  return 0;
+}
+
+static int
+process_filter_graph (GstFFMpegDeinterlace * deinterlace, AVPicture * dst,
+    const AVPicture * src, enum AVPixelFormat pixfmt, int width, int height)
+{
+  int res;
+
+  if (!deinterlace->filter_graph || width != deinterlace->last_width ||
+      height != deinterlace->last_height
+      || pixfmt != deinterlace->last_pixfmt) {
+    res = init_filter_graph (deinterlace, pixfmt, width, height);
+    if (res < 0)
+      return res;
+  }
+
+  memcpy (deinterlace->filter_frame->data, src->data, sizeof (src->data));
+  memcpy (deinterlace->filter_frame->linesize, src->linesize,
+      sizeof (src->linesize));
+  deinterlace->filter_frame->width = width;
+  deinterlace->filter_frame->height = height;
+  deinterlace->filter_frame->format = pixfmt;
+  res =
+      av_buffersrc_add_frame (deinterlace->buffersrc_ctx,
+      deinterlace->filter_frame);
+  if (res < 0)
+    return res;
+  res =
+      av_buffersink_get_frame (deinterlace->buffersink_ctx,
+      deinterlace->filter_frame);
+  if (res < 0)
+    return res;
+  av_picture_copy (dst, (const AVPicture *) deinterlace->filter_frame, pixfmt,
+      width, height);
+  av_frame_unref (deinterlace->filter_frame);
+
+  return 0;
 }
 
 static GstFlowReturn
@@ -320,8 +434,9 @@ gst_ffmpegdeinterlace_chain (GstPad * pad, GstObject * parent,
   gst_ffmpeg_avpicture_fill (&deinterlace->to_frame, to_map.data,
       deinterlace->pixfmt, deinterlace->width, deinterlace->height);
 
-  avpicture_deinterlace (&deinterlace->to_frame, &deinterlace->from_frame,
-      deinterlace->pixfmt, deinterlace->width, deinterlace->height);
+  process_filter_graph (deinterlace, &deinterlace->to_frame,
+      &deinterlace->from_frame, deinterlace->pixfmt, deinterlace->width,
+      deinterlace->height);
   gst_buffer_unmap (outbuf, &to_map);
   gst_buffer_unmap (inbuf, &from_map);
 
